part of image;

/**
 * A 32-bit image buffer where pixels are encoded into 32-bit unsigned ints.
 * You can use the methods in color to encode/decode the RGBA channels of a
 * color for a pixel.
 *
 * Pixels are stored in 32-bit unsigned integers in little-endian format:
 * aabbggrr.
 */
class Image {
  /// 24-bit RGB image.
  static const int RGB = 3;
  /// 32-bit RGBA image.
  static const int RGBA = 4;

  /// Width of the image.
  final int width;
  /// Height of the image.
  final int height;

  /**
   * Create an image with the given dimensions and format.
   */
  Image(int width, int height, [this._format = RGBA]) :
    this.width = width,
    this.height = height,
    _data = new Data.Uint32List(width * height);

  /**
   * Create a copy of the image [other].
   */
  Image.from(Image other) :
    width = other.width,
    height = other.height,
    _format = other._format,
    _data = new Data.Uint32List.fromList(other._data);

  /**
   * Create an image from [bytes].
   *
   * [bytes] should be in RGB<A> format with a byte [0,255] for each channel.
   * The length of [bytes] should be <3|4> * (width * height).
   * [format] determines if there are 3 or 4 channels per pixel.
   *
   * For example, given an Html Canvas, you could create an image:
   * var bytes = canvas.getContext('2d').getImageData(0, 0,
   *   canvas.width, canvas.height);
   * Image image = new Image.fromBytes(canvas.width, canvas.height, bytes);
   */
  Image.fromBytes(int width, int height, List<int> bytes,
                  [this._format = RGBA]) :
    this.width = width,
    this.height = height,
    // Create a uint32 view of the byte buffer.
    // This assumes the system architecture is little-endian...
    _data = new Data.Uint32List.view(new Data.Uint8List.fromList(bytes).buffer) {
  }

  /**
   * Clone this image.
   */
  Image clone() => new Image.from(this);

  /**
   * Get the RGBA bytes from the image.  You can use this to access the
   * RGBA color channels directly, or to pass it to something like an
   * Html canvas context.
   *
   * For example, given an Html Canvas, you could draw this image into the
   * canvas:
   * canvas.getContext('2d').putImageData(image.getBytes());
   */
  List<int> getBytes() =>
    new Data.Uint8List.view(_data.buffer);

  int get format => _format;

  /**
   * Change the format of the image.
   */
  void set format(int f) {
    if (f == _format) {
      return;
    }
    _format = f;
    if (_format == RGBA) {
      int len = _data.length;
      // If we convert from RGB to RGBA, make sure the alpha
      // channel is set.
      for (int i = 0; i < len; ++i) {
        _data[i] |= 0xff;
      }
    }
  }

  /**
   * How many color channels does the image have?
   */
  int get numChannels => _format;

  /**
   * Set all of the pixels of the image to the given [color].
   */
  Image fill(int color) {
    _data.fillRange(0, _data.length, color);
    return this;
  }

  /**
   * Add the colors of [other] to the pixels of this image.
   */
  Image operator+(Image other) {
    int h = Math.min(height, other.height);
    int w = Math.min(width, other.width);
    for (int y = 0; y < h; ++y) {
      for (int x = 0; x < w; ++x) {
        int c1 = getPixel(x, y);
        int r1 = getRed(c1);
        int g1 = getGreen(c1);
        int b1 = getBlue(c1);
        int a1 = getAlpha(c1);

        int c2 = other.getPixel(x, y);
        int r2 = getRed(c2);
        int g2 = getGreen(c2);
        int b2 = getBlue(c2);
        int a2 = getAlpha(c2);

        setPixel(x, y, getColor(r1 + r2, g1 + g2, b1 + b2, a1 + a2));
      }
    }
    return this;
  }

  /**
   * Subtract the colors of [other] from the pixels of this image.
   */
  Image operator-(Image other) {
    int h = Math.min(height, other.height);
    int w = Math.min(width, other.width);
    for (int y = 0; y < h; ++y) {
      for (int x = 0; x < w; ++x) {
        int c1 = getPixel(x, y);
        int r1 = getRed(c1);
        int g1 = getGreen(c1);
        int b1 = getBlue(c1);
        int a1 = getAlpha(c1);

        int c2 = other.getPixel(x, y);
        int r2 = getRed(c2);
        int g2 = getGreen(c2);
        int b2 = getBlue(c2);
        int a2 = getAlpha(c2);

        setPixel(x, y, getColor(r1 - r2, g1 - g2, b1 - b2, a1 - a2));
      }
    }
    return this;
  }

  /**
   * Multiply the colors of [other] with the pixels of this image.
   */
  Image operator*(Image other) {
    int h = Math.min(height, other.height);
    int w = Math.min(width, other.width);
    for (int y = 0; y < h; ++y) {
      for (int x = 0; x < w; ++x) {
        int c1 = getPixel(x, y);
        int r1 = getRed(c1);
        int g1 = getGreen(c1);
        int b1 = getBlue(c1);
        int a1 = getAlpha(c1);

        int c2 = other.getPixel(x, y);
        int r2 = getRed(c2);
        int g2 = getGreen(c2);
        int b2 = getBlue(c2);
        int a2 = getAlpha(c2);

        setPixel(x, y, getColor(r1 * r2, g1 * g2, b1 * b2, a1 * a2));
      }
    }
    return this;
  }

  /**
   * OR the colors of [other] to the pixels of this image.
   */
  Image operator|(Image other) {
    int h = Math.min(height, other.height);
    int w = Math.min(width, other.width);
    for (int y = 0; y < h; ++y) {
      for (int x = 0; x < w; ++x) {
        int c1 = getPixel(x, y);
        int r1 = getRed(c1);
        int g1 = getGreen(c1);
        int b1 = getBlue(c1);
        int a1 = getAlpha(c1);

        int c2 = other.getPixel(x, y);
        int r2 = getRed(c2);
        int g2 = getGreen(c2);
        int b2 = getBlue(c2);
        int a2 = getAlpha(c2);

        setPixel(x, y, getColor(r1 | r2, g1 | g2, b1 | b2, a1 | a2));
      }
    }
    return this;
  }

  /**
   * AND the colors of [other] with the pixels of this image.
   */
  Image operator&(Image other) {
    int h = Math.min(height, other.height);
    int w = Math.min(width, other.width);
    for (int y = 0; y < h; ++y) {
      for (int x = 0; x < w; ++x) {
        int c1 = getPixel(x, y);
        int r1 = getRed(c1);
        int g1 = getGreen(c1);
        int b1 = getBlue(c1);
        int a1 = getAlpha(c1);

        int c2 = other.getPixel(x, y);
        int r2 = getRed(c2);
        int g2 = getGreen(c2);
        int b2 = getBlue(c2);
        int a2 = getAlpha(c2);

        setPixel(x, y, getColor(r1 & r2, g1 & g2, b1 & b2, a1 & a2));
      }
    }
    return this;
  }

  /**
   * Modula the colors of [other] with the pixels of this image.
   */
  Image operator%(Image other) {
    int h = Math.min(height, other.height);
    int w = Math.min(width, other.width);
    for (int y = 0; y < h; ++y) {
      for (int x = 0; x < w; ++x) {
        int c1 = getPixel(x, y);
        int r1 = getRed(c1);
        int g1 = getGreen(c1);
        int b1 = getBlue(c1);
        int a1 = getAlpha(c1);

        int c2 = other.getPixel(x, y);
        int r2 = getRed(c2);
        int g2 = getGreen(c2);
        int b2 = getBlue(c2);
        int a2 = getAlpha(c2);

        setPixel(x, y, getColor(r1 % r2, g1 % g2, b1 % b2, a1 % a2));
      }
    }
    return this;
  }

  /**
   * The size of the image buffer.
   */
  int get length => _data.length;

  /**
   * Get a pixel from the buffer.
   */
  int operator[](int index) => _data[index];

  /**
   * Set a pixel in the buffer.
   */
  void operator[]=(int index, int color) {
    _data[index] = color;
  }

  /**
   * Get the buffer index for the [x], [y] pixel coordinates.
   */
  int index(int x, int y) => y * width + x;

  /**
   * Is the given pixel coordinates within the resolution of the image.
   */
  bool boundsSafe(int x, int y) {
    return x >= 0 && x < width && y >= 0 && y < height;
  }

  /**
   * Get the pixel from the given [x], [y] coordinate.
   */
  int getPixel(int x, int y) =>
    boundsSafe(x, y) ?
      _format == RGBA ?
      _data[y * width + x] :
      _data[y * width + x] | 0xff000000 : 0;

  /**
   * Get the pixel using the given [interpolation] type for non-integer pixel
   * coordinates.
   */
  int getPixelInterpolate(num fx, num fy, [int interpolation = LINEAR]) {
    if (interpolation == CUBIC) {
      return getPixelCubic(fx, fy);
    } else if (interpolation == LINEAR) {
      return getPixelLinear(fx, fy);
    }
    return getPixel(fx.toInt(), fy.toInt());
  }

  /**
   * Get the pixel using linear interpolation for non-integer pixel
   * coordinates.
   */
  int getPixelLinear(num fx, num fy) {
    int x = fx.toInt() - (fx >= 0 ? 0 : 1);
    int nx = x + 1;
    int y = fy.toInt() - (fy >= 0 ? 0 : 1);
    int ny = y + 1;
    double dx = fx - x;
    double dy = fy - y;

    int _linear(int Icc, int Inc, int Icn, int Inn) {
      return (Icc + dx * (Inc - Icc + dy * (Icc + Inn - Icn - Inc)) +
              dy * (Icn - Icc)).toInt();
    }

    int Icc = getPixel(x, y);
    int Inc = getPixel(nx, y);
    int Icn = getPixel(x, ny);
    int Inn = getPixel(nx, ny);

    return getColor(
        _linear(getRed(Icc), getRed(Inc), getRed(Icn), getRed(Inn)),
        _linear(getGreen(Icc), getGreen(Inc), getGreen(Icn), getGreen(Inn)),
        _linear(getBlue(Icc), getBlue(Inc), getBlue(Icn), getBlue(Inn)),
        _linear(getAlpha(Icc), getAlpha(Inc), getAlpha(Icn), getAlpha(Inn)));
  }

  /**
   * Get the pixel using cubic interpolation for non-integer pixel
   * coordinates.
   */
  int getPixelCubic(num fx, num fy) {
    int x = fx.toInt() - (fx >= 0.0 ? 0 : 1);
    int px = x - 1;
    int nx = x + 1;
    int ax = x + 2;
    int y = fy.toInt() - (fy >= 0.0 ? 0 : 1);
    int py = y - 1;
    int ny = y + 1;
    int ay = y + 2;

    double dx = fx - x;
    double dy = fy - y;

    double _cubic(double dx, num Ipp, num Icp, num Inp, num Iap) {
      return Icp + 0.5 * (dx * (-Ipp + Inp) +
             dx * dx * (2 * Ipp - 5 * Icp + 4 * Inp - Iap) +
             dx * dx * dx * (-Ipp + 3 * Icp - 3 * Inp + Iap));
    }

    int Ipp = getPixel(px, py);
    int Icp = getPixel(x, py);
    int Inp = getPixel(nx, py);
    int Iap = getPixel(ax, py);
    double Ip0 = _cubic(dx, getRed(Ipp), getRed(Icp), getRed(Inp), getRed(Iap));
    double Ip1 = _cubic(dx, getGreen(Ipp), getGreen(Icp), getGreen(Inp), getGreen(Iap));
    double Ip2 = _cubic(dx, getBlue(Ipp), getBlue(Icp), getBlue(Inp), getBlue(Iap));
    double Ip3 = _cubic(dx, getAlpha(Ipp), getAlpha(Icp), getAlpha(Inp), getAlpha(Iap));

    int Ipc = getPixel(px, y);
    int Icc = getPixel(x, y);
    int Inc = getPixel(nx, y);
    int Iac = getPixel(ax, y);
    double Ic0 = _cubic(dx, getRed(Ipc), getRed(Icc), getRed(Inc), getRed(Iac));
    double Ic1 = _cubic(dx, getGreen(Ipc), getGreen(Icc), getGreen(Inc), getGreen(Iac));
    double Ic2 = _cubic(dx, getBlue(Ipc), getBlue(Icc), getBlue(Inc), getBlue(Iac));
    double Ic3 = _cubic(dx, getAlpha(Ipc), getAlpha(Icc), getAlpha(Inc), getAlpha(Iac));

    int Ipn = getPixel(px, ny);
    int Icn = getPixel(x, ny);
    int Inn = getPixel(nx, ny);
    int Ian = getPixel(ax, ny);
    double In0 = _cubic(dx, getRed(Ipn), getRed(Icn), getRed(Inn), getRed(Ian));
    double In1 = _cubic(dx, getGreen(Ipn), getGreen(Icn), getGreen(Inn), getGreen(Ian));
    double In2 = _cubic(dx, getBlue(Ipn), getBlue(Icn), getBlue(Inn), getBlue(Ian));
    double In3 = _cubic(dx, getAlpha(Ipn), getAlpha(Icn), getAlpha(Inn), getAlpha(Ian));

    int Ipa = getPixel(px, ay);
    int Ica = getPixel(x, ay);
    int Ina = getPixel(nx, ay);
    int Iaa = getPixel(ax, ay);
    double Ia0 = _cubic(dx, getRed(Ipa), getRed(Ica), getRed(Ina), getRed(Iaa));
    double Ia1 = _cubic(dx, getGreen(Ipa), getGreen(Ica), getGreen(Ina), getGreen(Iaa));
    double Ia2 = _cubic(dx, getBlue(Ipa), getBlue(Ica), getBlue(Ina), getBlue(Iaa));
    double Ia3 = _cubic(dx, getAlpha(Ipa), getAlpha(Ica), getAlpha(Ina), getAlpha(Iaa));

    double c0 = _cubic(dy, Ip0, Ic0, In0, Ia0);
    double c1 = _cubic(dy, Ip1, Ic1, In1, Ia1);
    double c2 = _cubic(dy, Ip2, Ic2, In2, Ia2);
    double c3 = _cubic(dy, Ip3, Ic3, In3, Ia3);

    return getColor(c0.toInt(), c1.toInt(), c2.toInt(), c3.toInt());
  }

  /**
   * Set the pixel at the given [x], [y] coordinate to the [color].
   *
   * This simply replaces the existing color, it does not do any alpha
   * blending.  Use [drawPixel] for that.
   */
  void setPixel(int x, int y, int color) {
    if (boundsSafe(x, y)) {
      _data[y * width + x] = color;
    }
  }

  /**
   * Set the pixel at the given [x], [y] coordinate to the color
   * [r], [g], [b], [a].
   *
   * This simply replaces the existing color, it does not do any alpha
   * blending.  Use [drawPixel] for that.
   */
  void setPixelRGBA(int x, int y, int r, int g, int b, [int a = 0xff]) {
    if (boundsSafe(x, y)) {
      _data[y * width + x] = getColor(r, g, b, a);
    }
  }

  /// Format of the image.
  int _format;
  /// Pixels are encoded into 4-byte integers, where each byte is an RGBA
  /// channel.
  final Data.Uint32List _data;
}
